
import pytest
from unittest.mock import MagicMock, patch
import time

try:
    from aird.handlers.auth_handlers import check_login_rate_limit, _LOGIN_ATTEMPTS
    from aird.constants import LOGIN_RATE_LIMIT_ATTEMPTS, LOGIN_RATE_LIMIT_WINDOW
    AIRD_AVAILABLE = True
except ImportError:
    AIRD_AVAILABLE = False

@pytest.mark.skipif(not AIRD_AVAILABLE, reason="aird module not available")
class TestRateLimiting:
    def setup_method(self):
        if AIRD_AVAILABLE:
            _LOGIN_ATTEMPTS.clear()

    def test_rate_limit_counters(self):
        ip = "1.2.3.4"
        for _ in range(LOGIN_RATE_LIMIT_ATTEMPTS):
            assert check_login_rate_limit(ip) is True
        
        # Next one should fail
        assert check_login_rate_limit(ip) is False

    def test_rate_limit_expiry(self):
        ip = "5.6.7.8"
        # Use up quota
        for _ in range(LOGIN_RATE_LIMIT_ATTEMPTS):
            check_login_rate_limit(ip)
        
        assert check_login_rate_limit(ip) is False
        
        # Mock time to be in future
        # Note: We need to ensure we patch the time function used in the module
        future_time = time.time() + LOGIN_RATE_LIMIT_WINDOW + 10
        with patch('aird.handlers.auth_handlers.time.time', return_value=future_time):
            # Should work now because window expired
            assert check_login_rate_limit(ip) is True



"""Tests for aird/database/feature_flags.py"""

import pytest
import sqlite3
from unittest.mock import patch, MagicMock

from aird.database.feature_flags import (
    load_feature_flags,
    save_feature_flags,
    load_websocket_config,
    save_websocket_config,
    is_feature_enabled
)


@pytest.fixture
def db_conn():
    """Create an in-memory SQLite database for testing"""
    conn = sqlite3.connect(":memory:")
    conn.execute("CREATE TABLE feature_flags (key TEXT PRIMARY KEY, value INTEGER)")
    conn.execute("CREATE TABLE websocket_config (key TEXT PRIMARY KEY, value INTEGER)")
    conn.commit()
    yield conn
    conn.close()


class TestLoadFeatureFlags:
    """Tests for load_feature_flags function"""
    
    def test_load_empty_flags(self, db_conn):
        """Test loading when no flags exist"""
        result = load_feature_flags(db_conn)
        assert result == {}
    
    def test_load_existing_flags(self, db_conn):
        """Test loading existing feature flags"""
        db_conn.execute("INSERT INTO feature_flags (key, value) VALUES ('flag1', 1)")
        db_conn.execute("INSERT INTO feature_flags (key, value) VALUES ('flag2', 0)")
        db_conn.commit()
        
        result = load_feature_flags(db_conn)
        assert result == {'flag1': True, 'flag2': False}
    
    def test_load_flags_converts_to_bool(self, db_conn):
        """Test that values are converted to boolean"""
        db_conn.execute("INSERT INTO feature_flags (key, value) VALUES ('flag1', 5)")
        db_conn.commit()
        
        result = load_feature_flags(db_conn)
        assert result['flag1'] is True
    
    def test_load_flags_exception_returns_empty(self):
        """Test that exceptions return empty dict"""
        mock_conn = MagicMock()
        mock_conn.execute.side_effect = Exception("Database error")
        
        result = load_feature_flags(mock_conn)
        assert result == {}


class TestSaveFeatureFlags:
    """Tests for save_feature_flags function"""
    
    def test_save_new_flags(self, db_conn):
        """Test saving new feature flags"""
        flags = {'flag1': True, 'flag2': False}
        save_feature_flags(db_conn, flags)
        
        result = load_feature_flags(db_conn)
        assert result == flags
    
    def test_save_updates_existing_flags(self, db_conn):
        """Test that saving updates existing flags"""
        db_conn.execute("INSERT INTO feature_flags (key, value) VALUES ('flag1', 0)")
        db_conn.commit()
        
        save_feature_flags(db_conn, {'flag1': True})
        
        result = load_feature_flags(db_conn)
        assert result['flag1'] is True
    
    def test_save_converts_bool_to_int(self, db_conn):
        """Test that boolean values are converted to integers"""
        save_feature_flags(db_conn, {'flag1': True, 'flag2': False})
        
        cursor = db_conn.execute("SELECT key, value FROM feature_flags ORDER BY key")
        rows = cursor.fetchall()
        assert rows == [('flag1', 1), ('flag2', 0)]
    
    def test_save_flags_exception_handled(self):
        """Test that exceptions during save are handled"""
        mock_conn = MagicMock()
        mock_conn.__enter__ = MagicMock(side_effect=Exception("Database error"))
        
        # Should not raise exception
        save_feature_flags(mock_conn, {'flag1': True})


class TestLoadWebsocketConfig:
    """Tests for load_websocket_config function"""
    
    def test_load_empty_config(self, db_conn):
        """Test loading when no config exists"""
        result = load_websocket_config(db_conn)
        assert result == {}
    
    def test_load_existing_config(self, db_conn):
        """Test loading existing websocket config"""
        db_conn.execute("INSERT INTO websocket_config (key, value) VALUES ('max_connections', 100)")
        db_conn.execute("INSERT INTO websocket_config (key, value) VALUES ('timeout', 30)")
        db_conn.commit()
        
        result = load_websocket_config(db_conn)
        assert result == {'max_connections': 100, 'timeout': 30}
    
    def test_load_config_creates_table_if_missing(self):
        """Test that load creates table if it doesn't exist"""
        conn = sqlite3.connect(":memory:")
        result = load_websocket_config(conn)
        assert result == {}
        
        # Verify table was created
        cursor = conn.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='websocket_config'")
        assert cursor.fetchone() is not None
        conn.close()
    
    def test_load_config_exception_returns_empty(self):
        """Test that exceptions return empty dict"""
        mock_conn = MagicMock()
        mock_conn.execute.side_effect = Exception("Database error")
        
        result = load_websocket_config(mock_conn)
        assert result == {}


class TestSaveWebsocketConfig:
    """Tests for save_websocket_config function"""
    
    def test_save_new_config(self, db_conn):
        """Test saving new websocket config"""
        config = {'max_connections': 100, 'timeout': 30}
        save_websocket_config(db_conn, config)
        
        result = load_websocket_config(db_conn)
        assert result == config
    
    def test_save_updates_existing_config(self, db_conn):
        """Test that saving updates existing config"""
        db_conn.execute("INSERT INTO websocket_config (key, value) VALUES ('timeout', 10)")
        db_conn.commit()
        
        save_websocket_config(db_conn, {'timeout': 60})
        
        result = load_websocket_config(db_conn)
        assert result['timeout'] == 60
    
    def test_save_config_creates_table_if_missing(self):
        """Test that save creates table if it doesn't exist"""
        conn = sqlite3.connect(":memory:")
        save_websocket_config(conn, {'key1': 100})
        
        # Verify table was created and data was saved
        cursor = conn.execute("SELECT value FROM websocket_config WHERE key = 'key1'")
        assert cursor.fetchone()[0] == 100
        conn.close()
    
    def test_save_config_exception_handled(self):
        """Test that exceptions during save are handled"""
        mock_conn = MagicMock()
        mock_conn.execute.side_effect = Exception("Database error")
        
        # Should not raise exception
        save_websocket_config(mock_conn, {'key1': 100})


class TestIsFeatureEnabled:
    """Tests for is_feature_enabled function"""
    
    def test_feature_enabled_default_false(self):
        """Test default value is False"""
        with patch('aird.database.db.get_db_conn', return_value=None), \
             patch('aird.constants.FEATURE_FLAGS', {}):
            result = is_feature_enabled('nonexistent_feature')
            assert result is False
    
    def test_feature_enabled_custom_default(self):
        """Test custom default value"""
        with patch('aird.database.db.get_db_conn', return_value=None), \
             patch('aird.constants.FEATURE_FLAGS', {}):
            result = is_feature_enabled('nonexistent_feature', default=True)
            assert result is True
    
    def test_feature_enabled_from_constants(self):
        """Test reading from FEATURE_FLAGS constant"""
        with patch('aird.database.db.get_db_conn', return_value=None), \
             patch('aird.constants.FEATURE_FLAGS', {'my_feature': True}):
            result = is_feature_enabled('my_feature')
            assert result is True
    
    def test_feature_enabled_from_database(self):
        """Test reading from database"""
        conn = sqlite3.connect(":memory:")
        conn.execute("CREATE TABLE feature_flags (key TEXT PRIMARY KEY, value INTEGER)")
        conn.execute("INSERT INTO feature_flags (key, value) VALUES ('db_feature', 1)")
        conn.commit()
        
        with patch('aird.database.db.get_db_conn', return_value=conn), \
             patch('aird.constants.FEATURE_FLAGS', {}):
            result = is_feature_enabled('db_feature')
            assert result is True
        
        conn.close()
    
    def test_feature_database_overrides_constants(self):
        """Test that database values override constant values"""
        conn = sqlite3.connect(":memory:")
        conn.execute("CREATE TABLE feature_flags (key TEXT PRIMARY KEY, value INTEGER)")
        conn.execute("INSERT INTO feature_flags (key, value) VALUES ('my_feature', 0)")
        conn.commit()
        
        with patch('aird.database.db.get_db_conn', return_value=conn), \
             patch('aird.constants.FEATURE_FLAGS', {'my_feature': True}):
            result = is_feature_enabled('my_feature')
            assert result is False  # Database value (0/False) overrides constant (True)
        
        conn.close()
    
    def test_feature_enabled_exception_uses_default(self):
        """Test that exceptions fall back to default"""
        mock_conn = MagicMock()
        mock_conn.execute.side_effect = Exception("Database error")
        
        with patch('aird.database.db.get_db_conn', return_value=mock_conn), \
             patch('aird.constants.FEATURE_FLAGS', {'my_feature': True}):
            result = is_feature_enabled('my_feature')
            # Should use the value from FEATURE_FLAGS since DB failed
            assert result is True



"""User management database functions."""

import secrets
import json
import logging
import sqlite3
import hashlib
from datetime import datetime


# Import password hashing dependencies
try:
    from argon2 import PasswordHasher
    from argon2 import exceptions as argon2_exceptions
    ARGON2_AVAILABLE = True
    PH = PasswordHasher(time_cost=3, memory_cost=65536, parallelism=2)
except Exception:
    ARGON2_AVAILABLE = False
    PH = None


def hash_password(password: str) -> str:
    """Hash a password using Argon2 (Priority 1). Falls back to Scrypt if Argon2 unavailable."""
    if ARGON2_AVAILABLE and PH is not None:
        return PH.hash(password)
    # Fallback: Scrypt (better than SHA-256)
    salt = secrets.token_hex(16)
    key = hashlib.scrypt(password.encode('utf-8'), salt=salt.encode('utf-8'), n=16384, r=8, p=1, dklen=32)
    return f"scrypt:{salt}:{key.hex()}"


def verify_password(password: str, password_hash: str) -> bool:
    """Verify password supporting Argon2, Scrypt, and legacy salted SHA-256."""
    if not password_hash:
        return False

    # Try Argon2 first
    if password_hash.startswith("$argon2"):
        if ARGON2_AVAILABLE and PH is not None:
            try:
                return PH.verify(password_hash, password)
            except argon2_exceptions.VerifyMismatchError:
                return False
            except Exception:
                return False
        else:
            return False
            
    # Try Scrypt fallback
    if password_hash.startswith("scrypt:"):
        try:
            parts = password_hash.split(':')
            if len(parts) != 3:
                return False
            _, salt, stored_key_hex = parts
            key = hashlib.scrypt(password.encode('utf-8'), salt=salt.encode('utf-8'), n=16384, r=8, p=1, dklen=32)
            return secrets.compare_digest(key.hex(), stored_key_hex)
        except Exception:
            return False

    # Legacy format: salt:hash
    try:
        parts = password_hash.split(':', 1)
        if len(parts) != 2:
            return False
        salt, stored_hash = parts
        pwd_hash = hashlib.sha256((salt + password).encode('utf-8')).hexdigest()
        return secrets.compare_digest(pwd_hash, stored_hash)
    except Exception:
        return False


def create_user(conn: sqlite3.Connection, username: str, password: str, role: str = 'user') -> dict:
    """Create a new user in the database"""
    try:
        password_hash = hash_password(password)
        created_at = datetime.now().isoformat()
        
        with conn:
            cursor = conn.execute(
                "INSERT INTO users (username, password_hash, role, created_at) VALUES (?, ?, ?, ?)",
                (username, password_hash, role, created_at)
            )
            user_id = cursor.lastrowid
            
        return {
            "id": user_id,
            "username": username,
            "role": role,
            "created_at": created_at,
            "active": True,
            "last_login": None
        }
    except sqlite3.IntegrityError:
        raise ValueError(f"Username '{username}' already exists")
    except Exception as e:
        raise Exception(f"Failed to create user: {str(e)}")


def get_user_by_username(conn: sqlite3.Connection, username: str) -> dict | None:
    """Get user by username"""
    try:
        row = conn.execute(
            "SELECT id, username, password_hash, role, created_at, active, last_login FROM users WHERE username = ?",
            (username,)
        ).fetchone()
        
        if row:
            return {
                "id": row[0],
                "username": row[1],
                "password_hash": row[2],
                "role": row[3],
                "created_at": row[4],
                "active": bool(row[5]),
                "last_login": row[6]
            }
        return None
    except Exception:
        return None


def get_all_users(conn: sqlite3.Connection) -> list[dict]:
    """Get all users from the database"""
    try:
        rows = conn.execute(
            "SELECT id, username, role, created_at, active, last_login FROM users ORDER BY created_at DESC"
        ).fetchall()
        
        return [
            {
                "id": row[0],
                "username": row[1],
                "role": row[2],
                "created_at": row[3],
                "active": bool(row[4]),
                "last_login": row[5]
            }
            for row in rows
        ]
    except Exception:
        return []


def search_users(conn: sqlite3.Connection, query: str) -> list[dict]:
    """Search users by username (case-insensitive)"""
    try:
        rows = conn.execute(
            "SELECT id, username, role, created_at, active, last_login FROM users WHERE username LIKE ? AND active = 1 ORDER BY username LIMIT 20",
            (f"%{query}%",)
        ).fetchall()
        
        return [
            {
                "id": row[0],
                "username": row[1],
                "role": row[2],
                "created_at": row[3],
                "active": bool(row[4]),
                "last_login": row[5]
            }
            for row in rows
        ]
    except Exception:
        return []


def update_user(conn: sqlite3.Connection, user_id: int, **kwargs) -> bool:
    """Update user information"""
    try:
        valid_fields = ['username', 'password', 'password_hash', 'role', 'active', 'last_login']
        updates = []
        values = []
        
        for field, value in kwargs.items():
            if field in valid_fields:
                if field == 'password' and value:  # Special handling for password
                    updates.append('password_hash = ?')
                    values.append(hash_password(value))
                elif field == 'active':
                    updates.append('active = ?')
                    values.append(1 if value else 0)
                else:
                    updates.append(f'{field} = ?')
                    values.append(value)
        
        if not updates:
            return False
            
        values.append(user_id)
        query = f"UPDATE users SET {', '.join(updates)} WHERE id = ?"
        
        with conn:
            conn.execute(query, values)
        return True
    except Exception:
        return False


def delete_user(conn: sqlite3.Connection, user_id: int) -> bool:
    """Delete a user from the database"""
    try:
        with conn:
            cursor = conn.execute("DELETE FROM users WHERE id = ?", (user_id,))
            return cursor.rowcount > 0
    except Exception:
        return False


def authenticate_user(conn: sqlite3.Connection, username: str, password: str) -> dict | None:
    """Authenticate a user and return user data if successful"""
    user = get_user_by_username(conn, username)
    if not user or not user.get('active'):
        return None
    
    if verify_password(password, user['password_hash']):
        # Update last login time
        update_user(conn, user['id'], last_login=datetime.now().isoformat())
        return user
    
    return None


def assign_admin_privileges(conn: sqlite3.Connection, admin_users: list) -> None:
    """Assign admin role to specified users"""
    if not admin_users:
        return
    
    for username in admin_users:
        try:
            user = get_user_by_username(conn, username)
            if user and user['role'] != 'admin':
                update_user(conn, user['id'], role='admin')
                logging.info(f"Assigned admin privileges to user: {username}")
        except Exception as e:
            logging.error(f"Failed to assign admin privileges to {username}: {e}")




import pytest
from unittest.mock import patch, MagicMock, AsyncMock
from aird.handlers.file_op_handlers import (
    UploadHandler, DeleteHandler, RenameHandler, EditHandler, CloudUploadHandler
)
from aird.cloud import CloudProviderError
import json
import os
import asyncio

from tests.handler_helpers import authenticate, prepare_handler


class TestUploadHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}
        self.mock_request.headers = {
            'X-Upload-Dir': 'uploads',
            'X-Upload-Filename': 'test.txt'
        }

    def _setup_handler_for_post(self, handler, filename='test.txt', upload_dir='uploads'):
        """Helper to set up handler attributes typically set in prepare()"""
        handler._reject = False
        handler._reject_reason = None
        handler._temp_path = '/tmp/test_upload'
        handler._moved = False
        handler._too_large = False
        handler._writer_task = None
        handler._aiofile = AsyncMock()
        handler._buffer = []
        handler._writing = False
        handler._bytes_received = 100
        handler.upload_dir = upload_dir
        handler.filename = filename

    @pytest.mark.asyncio
    async def test_upload_success(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler)
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.realpath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('aird.handlers.file_op_handlers.ALLOWED_UPLOAD_EXTENSIONS', {'.txt'}), \
             patch('shutil.move') as mock_move, \
             patch('os.makedirs'), \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_move.assert_called()
            assert mock_write.call_args[0][0] == "Upload successful"

    @pytest.mark.asyncio
    async def test_upload_feature_disabled(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler)
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(403)
            assert "disabled" in mock_write.call_args[0][0].lower()

    @pytest.mark.asyncio
    async def test_upload_rejected_missing_header(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        
        # Simulate rejection in prepare() due to missing header
        handler._reject = True
        handler._reject_reason = "Missing X-Upload-Filename header"
        handler._temp_path = None
        handler._moved = False
        handler._too_large = False
        handler._writer_task = None
        handler._aiofile = None
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(400)
            assert "Missing X-Upload-Filename" in mock_write.call_args[0][0]

    @pytest.mark.asyncio
    async def test_upload_file_too_large(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler)
        handler._too_large = True
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(413)
            assert "too large" in mock_write.call_args[0][0].lower()

    @pytest.mark.asyncio
    async def test_upload_path_outside_root(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler, upload_dir='../../../etc')
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.realpath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(403)
            assert "denied" in mock_write.call_args[0][0].lower()

    @pytest.mark.asyncio
    async def test_upload_invalid_filename_dot(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler, filename='.')
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.realpath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.basename', return_value='.'), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(400)
            assert "invalid filename" in mock_write.call_args[0][0].lower()

    @pytest.mark.asyncio
    async def test_upload_invalid_filename_dotdot(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler, filename='..')
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.realpath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.basename', return_value='..'), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(400)
            assert "invalid filename" in mock_write.call_args[0][0].lower()

    @pytest.mark.asyncio
    async def test_upload_unsupported_file_type(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler, filename='malware.exe')
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.realpath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('aird.handlers.file_op_handlers.ALLOWED_UPLOAD_EXTENSIONS', {'.txt', '.pdf'}), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(415)
            assert "unsupported file type" in mock_write.call_args[0][0].lower()

    @pytest.mark.asyncio
    async def test_upload_filename_too_long(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        long_filename = 'a' * 260 + '.txt'
        self._setup_handler_for_post(handler, filename=long_filename)
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.realpath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('aird.handlers.file_op_handlers.ALLOWED_UPLOAD_EXTENSIONS', {'.txt'}), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(400)
            assert "too long" in mock_write.call_args[0][0].lower()

    @pytest.mark.asyncio
    async def test_upload_final_path_outside_root(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler)
        
        # First check passes, second fails (final path validation)
        is_within_root_calls = [True, False]
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.realpath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', side_effect=is_within_root_calls), \
             patch('aird.handlers.file_op_handlers.ALLOWED_UPLOAD_EXTENSIONS', {'.txt'}), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(403)
            assert "denied" in mock_write.call_args[0][0].lower()

    @pytest.mark.asyncio
    async def test_upload_move_failure(self):
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler)
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.realpath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('aird.handlers.file_op_handlers.ALLOWED_UPLOAD_EXTENSIONS', {'.txt'}), \
             patch('shutil.move', side_effect=OSError("Permission denied")), \
             patch('os.makedirs'), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            mock_set_status.assert_called_with(500)
            assert "failed to save" in mock_write.call_args[0][0].lower()

    @pytest.mark.asyncio
    async def test_upload_with_writer_task(self):
        """Test that upload waits for in-flight writes"""
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler)
        
        # Create a completed task
        async def dummy_task():
            pass
        handler._writer_task = asyncio.create_task(dummy_task())
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.realpath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('aird.handlers.file_op_handlers.ALLOWED_UPLOAD_EXTENSIONS', {'.txt'}), \
             patch('shutil.move'), \
             patch('os.makedirs'), \
             patch.object(handler, 'write') as mock_write:
            
            await handler.post()
            assert mock_write.call_args[0][0] == "Upload successful"

    def test_data_received_normal(self):
        handler = UploadHandler(self.mock_app, self.mock_request)
        handler._reject = False
        handler._bytes_received = 0
        handler._too_large = False
        handler._buffer = []
        handler._writing = False
        handler._aiofile = AsyncMock()
        
        with patch('aird.handlers.file_op_handlers.constants_module.MAX_FILE_SIZE', 1000), \
             patch('asyncio.create_task') as mock_create_task:
            handler.data_received(b'test data')
            
        assert handler._bytes_received == 9
        assert b'test data' in handler._buffer
        assert not handler._too_large
        mock_create_task.assert_called_once()

    def test_data_received_too_large(self):
        handler = UploadHandler(self.mock_app, self.mock_request)
        handler._reject = False
        handler._bytes_received = 900
        handler._too_large = False
        handler._buffer = []
        handler._writing = False
        
        with patch('aird.handlers.file_op_handlers.constants_module.MAX_FILE_SIZE', 1000):
            handler.data_received(b'x' * 200)  # Push over limit
            
        assert handler._too_large

    def test_data_received_rejected(self):
        handler = UploadHandler(self.mock_app, self.mock_request)
        handler._reject = True
        handler._bytes_received = 0
        handler._buffer = []
        
        handler.data_received(b'test data')
        
        # Should not process data when rejected
        assert handler._bytes_received == 0
        assert len(handler._buffer) == 0

    def test_on_finish_cleanup_temp_file(self):
        handler = UploadHandler(self.mock_app, self.mock_request)
        handler._temp_path = '/tmp/test_upload'
        handler._moved = False
        
        with patch('os.path.exists', return_value=True), \
             patch('os.remove') as mock_remove:
            handler.on_finish()
            mock_remove.assert_called_with('/tmp/test_upload')

    def test_on_finish_no_cleanup_when_moved(self):
        handler = UploadHandler(self.mock_app, self.mock_request)
        handler._temp_path = '/tmp/test_upload'
        handler._moved = True
        
        with patch('os.path.exists', return_value=True), \
             patch('os.remove') as mock_remove:
            handler.on_finish()
            mock_remove.assert_not_called()

    def test_on_finish_no_temp_path(self):
        handler = UploadHandler(self.mock_app, self.mock_request)
        handler._temp_path = None
        handler._moved = False
        
        # Should not raise
        handler.on_finish()

class TestUploadHandlerDynamicMaxSize:
    """Tests for dynamic max file size enforcement in UploadHandler"""

    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}
        self.mock_request.headers = {
            'X-Upload-Dir': 'uploads',
            'X-Upload-Filename': 'test.txt'
        }

    def _setup_handler_for_post(self, handler, filename='test.txt', upload_dir='uploads'):
        handler._reject = False
        handler._reject_reason = None
        handler._temp_path = '/tmp/test_upload'
        handler._moved = False
        handler._too_large = False
        handler._writer_task = None
        handler._aiofile = AsyncMock()
        handler._buffer = []
        handler._writing = False
        handler._bytes_received = 100
        handler.upload_dir = upload_dir
        handler.filename = filename

    def test_data_received_uses_dynamic_max_file_size(self):
        """Test data_received checks against constants_module.MAX_FILE_SIZE"""
        handler = UploadHandler(self.mock_app, self.mock_request)
        handler._reject = False
        handler._bytes_received = 0
        handler._too_large = False
        handler._buffer = []
        handler._writing = False

        custom_limit = 10 * 1024 * 1024  # 10 MB

        with patch('aird.handlers.file_op_handlers.constants_module') as mock_constants, \
             patch('asyncio.create_task'):
            mock_constants.MAX_FILE_SIZE = custom_limit

            small_chunk = b'x' * (5 * 1024 * 1024)
            handler.data_received(small_chunk)
            assert handler._too_large is False

            handler.data_received(small_chunk)
            handler.data_received(small_chunk)
            assert handler._too_large is True

    def test_data_received_respects_larger_limit(self):
        """Test data_received allows larger files when limit is increased"""
        handler = UploadHandler(self.mock_app, self.mock_request)
        handler._reject = False
        handler._bytes_received = 0
        handler._too_large = False
        handler._buffer = []
        handler._writing = False

        large_limit = 2 * 1024 * 1024 * 1024  # 2 GB

        with patch('aird.handlers.file_op_handlers.constants_module') as mock_constants, \
             patch('asyncio.create_task'):
            mock_constants.MAX_FILE_SIZE = large_limit

            chunk = b'x' * (500 * 1024 * 1024)
            handler.data_received(chunk)
            assert handler._too_large is False
            assert handler._bytes_received == 500 * 1024 * 1024

    @pytest.mark.asyncio
    async def test_too_large_error_shows_configured_limit(self):
        """Test that 413 error message reflects the configured limit"""
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler)
        handler._too_large = True

        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('aird.handlers.file_op_handlers.constants_module') as mock_constants, \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:

            mock_constants.UPLOAD_CONFIG = {'max_file_size_mb': 256}

            await handler.post()

            mock_set_status.assert_called_with(413)
            msg = mock_write.call_args[0][0]
            assert "256" in msg
            assert "too large" in msg.lower()

    @pytest.mark.asyncio
    async def test_too_large_error_shows_default_limit(self):
        """Test that 413 error shows 512 MB when using default config"""
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler)
        handler._too_large = True

        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('aird.handlers.file_op_handlers.constants_module') as mock_constants, \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:

            mock_constants.UPLOAD_CONFIG = {'max_file_size_mb': 512}

            await handler.post()

            mock_set_status.assert_called_with(413)
            msg = mock_write.call_args[0][0]
            assert "512" in msg

    @pytest.mark.asyncio
    async def test_too_large_error_shows_custom_large_limit(self):
        """Test that 413 error shows custom large limit (e.g. 2048 MB)"""
        handler = prepare_handler(UploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        self._setup_handler_for_post(handler)
        handler._too_large = True

        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('aird.handlers.file_op_handlers.constants_module') as mock_constants, \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:

            mock_constants.UPLOAD_CONFIG = {'max_file_size_mb': 2048}

            await handler.post()

            mock_set_status.assert_called_with(413)
            msg = mock_write.call_args[0][0]
            assert "2048" in msg


class TestDeleteHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}

    def test_delete_file(self):
        handler = prepare_handler(DeleteHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(return_value="test.txt")
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.abspath', return_value='/root/test.txt'), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.isfile', return_value=True), \
             patch('os.remove') as mock_remove, \
             patch.object(handler, 'redirect') as mock_redirect:
            
            handler.post()
            mock_remove.assert_called_with('/root/test.txt')
            mock_redirect.assert_called()

    def test_delete_feature_disabled(self):
        handler = prepare_handler(DeleteHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(403)
            assert "disabled" in mock_write.call_args[0][0].lower()

    def test_delete_access_denied(self):
        handler = prepare_handler(DeleteHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(return_value="../../../etc/passwd")
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.abspath', return_value='/etc/passwd'), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(403)
            assert "denied" in mock_write.call_args[0][0].lower()

    def test_delete_directory(self):
        handler = prepare_handler(DeleteHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(return_value="subdir")
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.abspath', return_value='/root/subdir'), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.isdir', return_value=True), \
             patch('os.path.isfile', return_value=False), \
             patch('shutil.rmtree') as mock_rmtree, \
             patch.object(handler, 'redirect') as mock_redirect:
            
            handler.post()
            mock_rmtree.assert_called_with('/root/subdir')
            mock_redirect.assert_called()

    def test_delete_with_parent_path(self):
        handler = prepare_handler(DeleteHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(return_value="subdir/file.txt")
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.abspath', return_value='/root/subdir/file.txt'), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.isfile', return_value=True), \
             patch('os.remove'), \
             patch.object(handler, 'redirect') as mock_redirect:
            
            handler.post()
            # Should redirect to parent directory
            mock_redirect.assert_called_with('/files/subdir')

class TestRenameHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}

    def test_rename_file(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'old.txt', 'new_name': 'new.txt'}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.abspath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.exists', return_value=True), \
             patch('os.rename') as mock_rename, \
             patch.object(handler, 'redirect') as mock_redirect:
            
            handler.post()
            mock_rename.assert_called()
            mock_redirect.assert_called()

    def test_rename_feature_disabled(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(403)
            assert "disabled" in mock_write.call_args[0][0].lower()

    def test_rename_empty_path(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': '', 'new_name': 'new.txt'}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(400)
            assert "required" in mock_write.call_args[0][0].lower()

    def test_rename_empty_new_name(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'old.txt', 'new_name': ''}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(400)
            assert "required" in mock_write.call_args[0][0].lower()

    def test_rename_invalid_filename_dot(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'old.txt', 'new_name': '.'}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(400)
            assert "invalid filename" in mock_write.call_args[0][0].lower()

    def test_rename_invalid_filename_dotdot(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'old.txt', 'new_name': '..'}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(400)
            assert "invalid filename" in mock_write.call_args[0][0].lower()

    def test_rename_invalid_filename_with_slash(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'old.txt', 'new_name': 'path/name'}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(400)
            assert "invalid filename" in mock_write.call_args[0][0].lower()

    def test_rename_invalid_filename_with_backslash(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'old.txt', 'new_name': 'path\\name'}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(400)
            assert "invalid filename" in mock_write.call_args[0][0].lower()

    def test_rename_filename_too_long(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        long_name = 'a' * 260
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'old.txt', 'new_name': long_name}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(400)
            assert "too long" in mock_write.call_args[0][0].lower()

    def test_rename_access_denied(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': '../etc/passwd', 'new_name': 'new.txt'}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.abspath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(403)
            assert "denied" in mock_write.call_args[0][0].lower()

    def test_rename_file_not_found(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'nonexistent.txt', 'new_name': 'new.txt'}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.abspath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.exists', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(404)
            assert "not found" in mock_write.call_args[0][0].lower()

    def test_rename_os_error(self):
        handler = prepare_handler(RenameHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'old.txt', 'new_name': 'new.txt'}.get(k, d))
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('os.path.abspath', side_effect=lambda p: p), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.exists', return_value=True), \
             patch('os.rename', side_effect=OSError("Permission denied")), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(500)
            assert "failed" in mock_write.call_args[0][0].lower()

class TestEditHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}
        self.mock_request.headers = {}

    def test_edit_file(self):
        handler = prepare_handler(EditHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'test.txt', 'content': 'new content'}.get(k, d))
        
        mock_resolved_path = MagicMock()
        mock_resolved_path.__str__.return_value = '/root/test.txt'
        mock_resolved_path.parents = [] # ROOT_DIR not in parents for success case
        
        mock_path = MagicMock()
        mock_path.resolve.return_value = mock_resolved_path

        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('pathlib.Path.absolute', return_value=mock_path), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.isfile', return_value=True), \
             patch('os.makedirs'), \
             patch('tempfile.NamedTemporaryFile') as mock_temp, \
             patch('os.replace') as mock_replace, \
             patch.object(handler, 'write') as mock_write:
            
            mock_temp.return_value.__enter__.return_value.name = '/tmp/temp'
            
            handler.post()
            mock_replace.assert_called()
            assert mock_write.call_args[0][0] == "File saved successfully."

    def test_edit_file_access_denied_json(self):
        handler = prepare_handler(EditHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.request.headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
        handler.request.body = json.dumps({'path': 'test.txt', 'content': 'new content'}).encode('utf-8')
        
        mock_resolved_path = MagicMock()
        mock_resolved_path.__str__.return_value = '/root/test.txt'
        
        mock_path = MagicMock()
        mock_path.resolve.return_value = mock_resolved_path

        # Simulate access denied by returning False from is_within_root
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('pathlib.Path.absolute', return_value=mock_path), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(403)
            # Should have written an error message
            mock_write.assert_called()

    def test_edit_feature_disabled(self):
        handler = prepare_handler(EditHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(403)
            assert "disabled" in mock_write.call_args[0][0].lower()

    def test_edit_invalid_json(self):
        handler = prepare_handler(EditHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.request.headers = {'Content-Type': 'application/json'}
        handler.request.body = b'{"invalid json'
        
        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(400)
            assert "invalid" in mock_write.call_args[0][0].lower()

    def test_edit_file_not_found(self):
        handler = prepare_handler(EditHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'nonexistent.txt', 'content': 'new content'}.get(k, d))
        
        mock_resolved_path = MagicMock()
        mock_resolved_path.__str__.return_value = '/root/nonexistent.txt'
        
        mock_path = MagicMock()
        mock_path.resolve.return_value = mock_resolved_path

        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('pathlib.Path.absolute', return_value=mock_path), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.isfile', return_value=False), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(404)
            assert "not found" in mock_write.call_args[0][0].lower()

    def test_edit_save_error(self):
        handler = prepare_handler(EditHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'path': 'test.txt', 'content': 'new content'}.get(k, d))
        
        mock_resolved_path = MagicMock()
        mock_resolved_path.__str__.return_value = '/root/test.txt'
        
        mock_path = MagicMock()
        mock_path.resolve.return_value = mock_resolved_path

        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('pathlib.Path.absolute', return_value=mock_path), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.isfile', return_value=True), \
             patch('os.makedirs'), \
             patch('tempfile.NamedTemporaryFile', side_effect=OSError("Disk full")), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            handler.post()
            mock_set_status.assert_called_with(500)
            assert "error" in mock_write.call_args[0][0].lower()

    def test_edit_json_response(self):
        handler = prepare_handler(EditHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.request.headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
        handler.request.body = json.dumps({'path': 'test.txt', 'content': 'new content'}).encode('utf-8')
        
        mock_resolved_path = MagicMock()
        mock_resolved_path.__str__.return_value = '/root/test.txt'
        
        mock_path = MagicMock()
        mock_path.resolve.return_value = mock_resolved_path

        with patch('aird.handlers.file_op_handlers.is_feature_enabled', return_value=True), \
             patch('pathlib.Path.absolute', return_value=mock_path), \
             patch('aird.handlers.file_op_handlers.is_within_root', return_value=True), \
             patch('os.path.isfile', return_value=True), \
             patch('os.makedirs'), \
             patch('tempfile.NamedTemporaryFile') as mock_temp, \
             patch('os.replace'), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            
            mock_temp.return_value.__enter__.return_value.name = '/tmp/temp'
            
            handler.post()
            mock_set_status.assert_called_with(200)
            # Should return JSON response
            assert mock_write.call_args[0][0] == {"ok": True}

class TestCloudUploadHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}
        self.mock_request.files = {'file': [{'body': b'content', 'filename': 'test.txt'}]}

    @pytest.mark.asyncio
    async def test_cloud_upload(self):
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        
        mock_provider = MagicMock()
        mock_provider.upload_file.return_value = MagicMock(to_dict=lambda: {'id': '123'})
        mock_manager = MagicMock()
        mock_manager.get.return_value = mock_provider
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch.object(handler, 'write') as mock_write:
            await handler.post("provider1")
            mock_write.assert_called()
            assert mock_write.call_args[0][0]['file']['id'] == '123'

    @pytest.mark.asyncio
    async def test_cloud_upload_provider_not_found(self):
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        
        mock_manager = MagicMock()
        mock_manager.get.return_value = None
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            await handler.post("nonexistent_provider")
            mock_set_status.assert_called_with(404)
            assert "not configured" in str(mock_write.call_args[0][0]).lower()

    @pytest.mark.asyncio
    async def test_cloud_upload_no_file(self):
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.request.files = {}  # No files
        
        mock_provider = MagicMock()
        mock_manager = MagicMock()
        mock_manager.get.return_value = mock_provider
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            await handler.post("provider1")
            mock_set_status.assert_called_with(400)
            assert "no file" in str(mock_write.call_args[0][0]).lower()

    @pytest.mark.asyncio
    async def test_cloud_upload_file_too_large(self):
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        
        # Create a large file body
        large_body = b'x' * (600 * 1024 * 1024)  # 600MB (over 512MB limit)
        handler.request.files = {'file': [{'body': large_body, 'filename': 'large.bin'}]}
        
        mock_provider = MagicMock()
        mock_manager = MagicMock()
        mock_manager.get.return_value = mock_provider
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch('aird.handlers.file_op_handlers.constants_module.MAX_FILE_SIZE', 512 * 1024 * 1024), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            await handler.post("provider1")
            mock_set_status.assert_called_with(413)
            assert "too large" in str(mock_write.call_args[0][0]).lower()

    @pytest.mark.asyncio
    async def test_cloud_upload_provider_error(self):
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        
        mock_provider = MagicMock()
        mock_provider.upload_file.side_effect = CloudProviderError("Upload failed")
        mock_manager = MagicMock()
        mock_manager.get.return_value = mock_provider
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch('asyncio.to_thread', side_effect=CloudProviderError("Upload failed")), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            await handler.post("provider1")
            mock_set_status.assert_called_with(400)
            assert "upload failed" in str(mock_write.call_args[0][0]).lower()

    @pytest.mark.asyncio
    async def test_cloud_upload_generic_error(self):
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        
        mock_provider = MagicMock()
        mock_manager = MagicMock()
        mock_manager.get.return_value = mock_provider
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch('asyncio.to_thread', side_effect=Exception("Unknown error")), \
             patch.object(handler, 'set_status') as mock_set_status, \
             patch.object(handler, 'write') as mock_write:
            await handler.post("provider1")
            mock_set_status.assert_called_with(500)
            assert "failed" in str(mock_write.call_args[0][0]).lower()

    @pytest.mark.asyncio
    async def test_cloud_upload_with_parent_id(self):
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.get_body_argument = MagicMock(return_value="parent_folder_123")
        
        mock_cloud_file = MagicMock()
        mock_cloud_file.to_dict.return_value = {'id': '456', 'name': 'test.txt'}
        
        mock_provider = MagicMock()
        mock_provider.upload_file.return_value = mock_cloud_file
        mock_manager = MagicMock()
        mock_manager.get.return_value = mock_provider
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch('asyncio.to_thread', return_value=mock_cloud_file), \
             patch.object(handler, 'write') as mock_write:
            await handler.post("provider1")
            mock_write.assert_called()
            assert mock_write.call_args[0][0]['file']['id'] == '456'

    @pytest.mark.asyncio
    async def test_cloud_upload_empty_file(self):
        """Test uploading an empty file is allowed"""
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.request.files = {'file': [{'body': b'', 'filename': 'empty.txt'}]}
        
        mock_cloud_file = MagicMock()
        mock_cloud_file.to_dict.return_value = {'id': '789', 'name': 'empty.txt'}
        
        mock_provider = MagicMock()
        mock_provider.upload_file.return_value = mock_cloud_file
        mock_manager = MagicMock()
        mock_manager.get.return_value = mock_provider
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch('asyncio.to_thread', return_value=mock_cloud_file), \
             patch.object(handler, 'write') as mock_write:
            await handler.post("provider1")
            mock_write.assert_called()
            assert mock_write.call_args[0][0]['file']['id'] == '789'

    @pytest.mark.asyncio
    async def test_cloud_upload_sanitize_filename(self):
        """Test that filenames are sanitized"""
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.request.files = {'file': [{'body': b'content', 'filename': '../../../etc/passwd'}]}
        
        mock_cloud_file = MagicMock()
        mock_cloud_file.to_dict.return_value = {'id': '123', 'name': 'passwd'}
        
        mock_provider = MagicMock()
        mock_provider.upload_file.return_value = mock_cloud_file
        mock_manager = MagicMock()
        mock_manager.get.return_value = mock_provider
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch('asyncio.to_thread', return_value=mock_cloud_file) as mock_to_thread, \
             patch('aird.handlers.file_op_handlers.sanitize_cloud_filename', return_value='passwd') as mock_sanitize, \
             patch.object(handler, 'write'):
            await handler.post("provider1")
            # Verify sanitize was called
            mock_sanitize.assert_called_once()

    @pytest.mark.asyncio
    async def test_cloud_upload_default_filename(self):
        """Test default filename when none provided"""
        handler = prepare_handler(CloudUploadHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        handler.request.files = {'file': [{'body': b'content', 'filename': None}]}
        
        mock_cloud_file = MagicMock()
        mock_cloud_file.to_dict.return_value = {'id': '123', 'name': 'upload.bin'}
        
        mock_provider = MagicMock()
        mock_provider.upload_file.return_value = mock_cloud_file
        mock_manager = MagicMock()
        mock_manager.get.return_value = mock_provider
        self.mock_app.settings['cloud_manager'] = mock_manager
        
        with patch('asyncio.to_thread', return_value=mock_cloud_file), \
             patch('aird.handlers.file_op_handlers.sanitize_cloud_filename', return_value='upload.bin'), \
             patch.object(handler, 'write') as mock_write:
            await handler.post("provider1")
            mock_write.assert_called()




import pytest
from unittest.mock import patch, MagicMock
from aird.handlers.base_handler import BaseHandler
import json

from tests.handler_helpers import authenticate, patch_db_conn, prepare_handler

class TestBaseHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}
        self.mock_request.headers = {}
        self.mock_request.protocol = "http"

    def test_get_current_user_from_cookie(self):
        handler = prepare_handler(BaseHandler(self.mock_app, self.mock_request))
        user_data = {'username': 'testuser', 'role': 'user'}
        
        db_conn = MagicMock()
        with patch.object(handler, 'get_secure_cookie', return_value=json.dumps(user_data).encode('utf-8')), \
             patch_db_conn(db_conn, modules=['aird.handlers.base_handler']), \
             patch('aird.db.get_user_by_username', return_value=user_data):
            
            user = handler.get_current_user()
            assert user['username'] == 'testuser'

    def test_get_current_user_from_token_header(self):
        handler = prepare_handler(BaseHandler(self.mock_app, self.mock_request))
        handler.request.headers = {'Authorization': 'Bearer valid_token'}
        
        with patch.object(handler, 'get_secure_cookie', return_value=None), \
             patch('aird.handlers.base_handler.config_module.ACCESS_TOKEN', 'valid_token'):
            
            user = handler.get_current_user()
            assert user['username'] == 'token_user'
            assert user['role'] == 'admin'

    def test_is_admin_user_true(self):
        handler = prepare_handler(BaseHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='admin')
        assert handler.is_admin_user() is True

    def test_is_admin_user_false(self):
        handler = prepare_handler(BaseHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        with patch.object(handler, 'get_secure_cookie', return_value=None):
            assert handler.is_admin_user() is False

    def test_get_display_username(self):
        handler = prepare_handler(BaseHandler(self.mock_app, self.mock_request))
        
        # Test regular user
        with patch.object(handler, 'get_current_user', return_value={'username': 'user', 'role': 'user'}):
            assert handler.get_display_username() == 'user (User)'
        
        # Test admin user
        with patch.object(handler, 'get_current_user', return_value={'username': 'admin', 'role': 'admin'}):
            assert handler.get_display_username() == 'admin (Admin)'
        
        # Test token user
        with patch.object(handler, 'get_current_user', return_value={'username': 'token_user', 'role': 'admin'}):
            assert handler.get_display_username() == 'Admin (Token)'

    def test_csp_nonce_generation(self):
        """Test that CSP nonce is generated and is cryptographically strong."""
        handler = prepare_handler(BaseHandler(self.mock_app, self.mock_request))
        handler.prepare()
        
        nonce = handler.get_csp_nonce()
        assert nonce is not None
        assert len(nonce) > 16  # Base64 encoded 16 bytes should be ~24 chars
        
        # Nonce should be consistent for same request
        assert handler.get_csp_nonce() == nonce

    def test_csp_nonce_in_template_namespace(self):
        """Test that CSP nonce is available in template namespace."""
        handler = prepare_handler(BaseHandler(self.mock_app, self.mock_request))
        handler.prepare()
        
        namespace = handler.get_template_namespace()
        assert 'csp_nonce' in namespace
        assert namespace['csp_nonce'] == handler.get_csp_nonce()

    def test_csp_header_contains_nonce(self):
        """Test that CSP header contains nonce and not unsafe-inline for scripts."""
        handler = prepare_handler(BaseHandler(self.mock_app, self.mock_request))
        handler.prepare()
        handler._set_csp_header()
        
        # Get the set header
        calls = handler.set_header.call_args_list
        csp_calls = [c for c in calls if c[0][0] == 'Content-Security-Policy']
        assert len(csp_calls) > 0
        
        csp_value = csp_calls[-1][0][1]
        nonce = handler.get_csp_nonce()
        
        assert f"'nonce-{nonce}'" in csp_value
        assert "'unsafe-inline'" not in csp_value.split('script-src')[1].split(';')[0]




"""Memory-mapped file handler for efficient large file operations."""

import os
import mmap
from datetime import datetime
from aird.constants import MMAP_MIN_SIZE, CHUNK_SIZE


class MMapFileHandler:
    """Efficient file handling using memory mapping for large files"""
    
    @staticmethod
    def should_use_mmap(file_size: int) -> bool:
        """Determine if mmap should be used based on file size"""
        return file_size >= MMAP_MIN_SIZE
    
    @staticmethod
    async def serve_file_chunk(file_path: str, start: int = 0, end: int = None, chunk_size: int = CHUNK_SIZE):
        """Serve file chunks using mmap for efficient memory usage"""
        try:
            file_size = os.path.getsize(file_path)
            
            if not MMapFileHandler.should_use_mmap(file_size):
                # Use traditional method for small files
                with open(file_path, 'rb') as f:
                    f.seek(start)
                    remaining = (end - start + 1) if end is not None else file_size - start
                    while remaining > 0:
                        chunk = f.read(min(chunk_size, remaining))
                        if not chunk:
                            break
                        yield chunk
                        remaining -= len(chunk)
                return
            
            # Use mmap for large files
            with open(file_path, 'rb') as f:
                with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mm:
                    actual_end = min(end or file_size - 1, file_size - 1)
                    current = start
                    
                    while current <= actual_end:
                        chunk_end = min(current + chunk_size, actual_end + 1)
                        yield mm[current:chunk_end]
                        current = chunk_end
                        
        except (OSError, ValueError) as e:
            # Fallback to traditional method on mmap errors
            with open(file_path, 'rb') as f:
                f.seek(start)
                remaining = (end - start + 1) if end is not None else file_size - start
                while remaining > 0:
                    chunk = f.read(min(chunk_size, remaining))
                    if not chunk:
                        break
                    yield chunk
                    remaining -= len(chunk)
    
    @staticmethod
    def find_line_offsets(file_path: str, max_lines: int = None) -> list[int]:
        """Efficiently find line start offsets using mmap"""
        try:
            file_size = os.path.getsize(file_path)
            if not MMapFileHandler.should_use_mmap(file_size):
                # Use traditional method for small files
                offsets = [0]
                with open(file_path, 'rb') as f:
                    pos = 0
                    for line in f:
                        pos += len(line)
                        offsets.append(pos)
                        if max_lines and len(offsets) > max_lines:
                            break
                return offsets[:-1]  # Remove the last offset (EOF)
            
            # Use mmap for large files
            offsets = [0]
            with open(file_path, 'rb') as f:
                with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mm:
                    pos = 0
                    while pos < len(mm):
                        newline_pos = mm.find(b'\n', pos)
                        if newline_pos == -1:
                            break
                        pos = newline_pos + 1
                        offsets.append(pos)
                        if max_lines and len(offsets) > max_lines:
                            break
            return offsets[:-1]
            
        except (OSError, ValueError):
            # Fallback to traditional method
            offsets = [0]
            with open(file_path, 'rb') as f:
                pos = 0
                for line in f:
                    pos += len(line)
                    offsets.append(pos)
                    if max_lines and len(offsets) > max_lines:
                        break
            return offsets[:-1]
    
    @staticmethod
    def search_in_file(file_path: str, search_term: str, max_results: int = 100) -> list[dict]:
        """Efficiently search for text in file using mmap"""
        results = []
        try:
            file_size = os.path.getsize(file_path)
            search_bytes = search_term.encode('utf-8')
            
            if not MMapFileHandler.should_use_mmap(file_size):
                # Use traditional method for small files
                with open(file_path, 'r', encoding='utf-8', errors='replace') as f:
                    for line_num, line in enumerate(f, 1):
                        if search_term in line:
                            results.append({
                                "line_number": line_num,
                                "line_content": line.rstrip('\n'),
                                "match_positions": [i for i in range(len(line)) if line[i:].startswith(search_term)]
                            })
                            if len(results) >= max_results:
                                break
                return results
            
            # Use mmap for large files
            with open(file_path, 'rb') as f:
                with mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mm:
                    current_pos = 0
                    line_number = 1
                    line_start = 0
                    
                    while current_pos < len(mm) and len(results) < max_results:
                        newline_pos = mm.find(b'\n', current_pos)
                        if newline_pos == -1:
                            # Last line
                            line_bytes = mm[current_pos:]
                            if search_bytes in line_bytes:
                                line_content = line_bytes.decode('utf-8', errors='replace')
                                match_positions = []
                                start_pos = 0
                                while True:
                                    pos = line_content.find(search_term, start_pos)
                                    if pos == -1:
                                        break
                                    match_positions.append(pos)
                                    start_pos = pos + 1
                                results.append({
                                    "line_number": line_number,
                                    "line_content": line_content,
                                    "match_positions": match_positions
                                })
                            break
                        
                        line_bytes = mm[current_pos:newline_pos]
                        if search_bytes in line_bytes:
                            line_content = line_bytes.decode('utf-8', errors='replace')
                            match_positions = []
                            start_pos = 0
                            while True:
                                pos = line_content.find(search_term, start_pos)
                                if pos == -1:
                                    break
                                match_positions.append(pos)
                                start_pos = pos + 1
                            results.append({
                                "line_number": line_number,
                                "line_content": line_content,
                                "match_positions": match_positions
                            })
                        
                        current_pos = newline_pos + 1
                        line_number += 1
                        
        except (OSError, UnicodeDecodeError):
            # Fallback to traditional search
            with open(file_path, 'r', encoding='utf-8', errors='replace') as f:
                for line_num, line in enumerate(f, 1):
                    if search_term in line:
                        results.append({
                            "line_number": line_num,
                            "line_content": line.rstrip('\n'),
                            "match_positions": [i for i in range(len(line)) if line[i:].startswith(search_term)]
                        })
                        if len(results) >= max_results:
                            break
        
        return results


def get_files_in_directory(path="."):
    """Get all files in a directory with metadata."""
    files = []
    for entry in os.scandir(path):
        stat = entry.stat()
        files.append({
            "name": entry.name,
            "is_dir": entry.is_dir(),
            "size_bytes": stat.st_size,
            "size_str": f"{stat.st_size / 1024:.2f} KB" if not entry.is_dir() else "-",
            "modified": datetime.fromtimestamp(stat.st_mtime).strftime('%Y-%m-%d %H:%M:%S'),
            "modified_timestamp": int(stat.st_mtime)
        })
    return files


def is_video_file(filename):
    """Check if file is a supported video format"""
    ext = os.path.splitext(filename)[1].lower()
    video_extensions = {'.mp4', '.avi', '.mkv', '.mov', '.wmv', '.flv', '.webm', '.m4v', '.3gp', '.ogv'}
    return ext in video_extensions


def is_audio_file(filename):
    """Check if file is a supported audio format"""
    ext = os.path.splitext(filename)[1].lower()
    audio_extensions = {'.mp3', '.wav', '.flac', '.aac', '.ogg', '.m4a', '.wma'}
    return ext in audio_extensions




import pytest
from unittest.mock import patch, MagicMock
from aird.handlers.auth_handlers import LoginHandler, LogoutHandler, AdminLoginHandler, ProfileHandler, LDAPLoginHandler
import json

from tests.handler_helpers import authenticate, patch_db_conn, prepare_handler

class TestLoginHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {
            'cookie_secret': 'test_secret',
            'login_url': '/login',
            'debug': False
        }
        self.mock_request.connection = MagicMock()
        self.mock_request.protocol = "http"

    def test_get_login_page(self):
        handler = prepare_handler(LoginHandler(self.mock_app, self.mock_request))
        with patch.object(handler, 'render') as mock_render:
            handler.get()
            mock_render.assert_called_once()
            args, kwargs = mock_render.call_args
            assert args[0] == "login.html"

    def test_get_redirects_if_authenticated(self):
        handler = prepare_handler(LoginHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        with patch.object(handler, 'redirect') as mock_redirect:
            handler.get()
            mock_redirect.assert_called_once()

    def test_post_login_success(self):
        handler = prepare_handler(LoginHandler(self.mock_app, self.mock_request))
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'username': 'user', 'password': 'password'}.get(k, d))
        
        with patch_db_conn(MagicMock(), modules=['aird.handlers.auth_handlers']), \
             patch('aird.handlers.auth_handlers.authenticate_user', return_value={'username': 'user', 'role': 'user'}), \
             patch.object(handler, 'set_secure_cookie') as mock_cookie, \
             patch.object(handler, 'redirect') as mock_redirect:
            
            handler.post()
            assert mock_cookie.call_count == 2 # user and role
            mock_redirect.assert_called_with('/files/')

    def test_post_login_failure(self):
        handler = prepare_handler(LoginHandler(self.mock_app, self.mock_request))
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'username': 'user', 'password': 'wrong'}.get(k, d))
        
        with patch_db_conn(MagicMock(), modules=['aird.handlers.auth_handlers']), \
             patch('aird.handlers.auth_handlers.authenticate_user', return_value=None), \
             patch.object(handler, 'render') as mock_render:
            
            handler.post()
            mock_render.assert_called()
            assert mock_render.call_args[1]['error'] == "Invalid username or password."

class TestLogoutHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}
    
    def test_logout(self):
        handler = prepare_handler(LogoutHandler(self.mock_app, self.mock_request))
        with patch.object(handler, 'clear_cookie') as mock_clear, \
             patch.object(handler, 'redirect') as mock_redirect:
            handler.get()
            assert mock_clear.call_count == 2 # user and admin
            mock_redirect.assert_called_with('/login')

class TestAdminLoginHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}
        self.mock_request.protocol = "http"

    def test_get_admin_login(self):
        handler = prepare_handler(AdminLoginHandler(self.mock_app, self.mock_request))
        with patch.object(handler, 'is_admin_user', return_value=False), \
             patch.object(handler, 'render') as mock_render:
            handler.get()
            mock_render.assert_called_with("admin_login.html", error=None)

    def test_post_admin_login_success(self):
        handler = prepare_handler(AdminLoginHandler(self.mock_app, self.mock_request))
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'username': 'admin', 'password': 'password'}.get(k, d))
        
        with patch_db_conn(MagicMock(), modules=['aird.handlers.auth_handlers']), \
             patch('aird.handlers.auth_handlers.authenticate_user', return_value={'username': 'admin', 'role': 'admin'}), \
             patch.object(handler, 'set_secure_cookie') as mock_cookie, \
             patch.object(handler, 'redirect') as mock_redirect:
            
            handler.post()
            assert mock_cookie.call_count == 3 # user, role, admin
            mock_redirect.assert_called_with('/admin')

class TestProfileHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {'cookie_secret': 'test_secret'}
        self.mock_request.protocol = "http"

    def test_get_profile(self):
        handler = prepare_handler(ProfileHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user')
        with patch.object(handler, 'render') as mock_render:
            handler.get()
            mock_render.assert_called()
            assert mock_render.call_args[0][0] == "profile.html"

    def test_post_update_password(self):
        handler = prepare_handler(ProfileHandler(self.mock_app, self.mock_request))
        authenticate(handler, role='user', username='user')
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'new_password': 'newpass1', 'confirm_password': 'newpass1'}.get(k, d))
        
        with patch_db_conn(MagicMock(), modules=['aird.handlers.auth_handlers']), \
             patch('aird.handlers.auth_handlers.get_user_by_username', return_value={'id': 1, 'username': 'user'}), \
             patch('aird.handlers.auth_handlers.update_user') as mock_update, \
             patch.object(handler, 'render') as mock_render:
            
            handler.post()
            mock_update.assert_called()
            assert mock_render.call_args[1]['success'] == "Password updated successfully"

class TestLDAPLoginHandler:
    def setup_method(self):
        self.mock_app = MagicMock()
        self.mock_request = MagicMock()
        self.mock_app.settings = {
            'cookie_secret': 'test_secret',
            'ldap_server': 'ldap://localhost',
            'ldap_user_template': 'uid={username},ou=users,dc=example,dc=com',
            'ldap_base_dn': 'ou=users,dc=example,dc=com',
            'ldap_filter_template': '(uid={username})',
            'ldap_attributes': ['cn', 'mail', 'uid'],
            'ldap_attribute_map': []
        }
        self.mock_request.protocol = "http"

    def test_get_ldap_login(self):
        handler = prepare_handler(LDAPLoginHandler(self.mock_app, self.mock_request))
        with patch.object(handler, 'render') as mock_render:
            handler.get()
            mock_render.assert_called_with("login.html", error=None, settings=self.mock_app.settings)

    def test_post_ldap_login_success(self):
        handler = prepare_handler(LDAPLoginHandler(self.mock_app, self.mock_request))
        handler.get_argument = MagicMock(side_effect=lambda k, d=None: {'username': 'user', 'password': 'password'}.get(k, d))
        
        mock_conn = MagicMock()
        mock_conn.search.return_value = True
        mock_conn.entries = [MagicMock(entry_attributes_as_dict={'uid': ['user'], 'mail': ['user@example.com'], 'cn': ['User Name']})]
        
        with patch('aird.handlers.auth_handlers.Server'), \
             patch('aird.handlers.auth_handlers.Connection', return_value=mock_conn), \
             patch_db_conn(MagicMock(), modules=['aird.handlers.auth_handlers']), \
             patch('aird.handlers.auth_handlers.get_user_by_username', return_value=None), \
             patch('aird.handlers.auth_handlers.create_user') as mock_create, \
             patch.object(handler, 'set_secure_cookie') as mock_cookie, \
             patch.object(handler, 'redirect') as mock_redirect:
            
            handler.post()
            mock_create.assert_called()
            mock_cookie.assert_called()
            mock_redirect.assert_called_with('/files/')



"""
Shared helpers for Tornado handler unit tests.
"""

from contextlib import contextmanager
from unittest.mock import MagicMock, patch


def _ensure_mock(attr_name, handler):
    attr = getattr(handler, attr_name, None)
    if not isinstance(attr, MagicMock):
        setattr(handler, attr_name, MagicMock())


def prepare_handler(handler):
    """Ensure minimal Tornado handler internals exist for isolated testing."""
    if not getattr(handler, "_transforms", None):
        handler._transforms = []
    handler._write_buffer = getattr(handler, "_write_buffer", [])
    handler._headers_written = getattr(handler, "_headers_written", False)
    if handler.request is None:
        handler.request = MagicMock()
    if not getattr(handler.request, "connection", None):
        handler.request.connection = MagicMock()
    if not getattr(handler.request.connection, "context", None):
        handler.request.connection.context = MagicMock()
    if not getattr(handler.request, "headers", None):
        handler.request.headers = {}
    if not getattr(handler.request, "path", None):
        handler.request.path = "/test"
    if not getattr(handler.request, "protocol", None):
        handler.request.protocol = "http"

    # Default mocks for common methods so tests can assert calls safely
    for method_name in (
        "write",
        "set_status",
        "render",
        "redirect",
        "finish",
        "flush",
        "clear_cookie",
        "set_header",
    ):
        _ensure_mock(method_name, handler)

    if not getattr(handler, "get_argument", None):
        handler.get_argument = MagicMock()
    if not getattr(handler, "get_secure_cookie", None):
        handler.get_secure_cookie = MagicMock(return_value=None)
    if not getattr(handler, "set_secure_cookie", None):
        handler.set_secure_cookie = MagicMock()

    return handler


def authenticate(handler, role="admin", username=None):
    """Stub get_current_user to satisfy @authenticated decorators."""
    prepare_handler(handler)
    username = username or ("admin" if role == "admin" else "user")
    user = {"username": username, "role": role}
    handler._current_user = user
    handler.get_current_user = MagicMock(return_value=user)
    return user


@contextmanager
def patch_db_conn(value, modules=None):
    """
    Patch DB_CONN in aird.constants and common handler modules.
    Extra module paths can be provided via `modules`.
    """
    targets = ["aird.constants.DB_CONN"]
    default_modules = [
        "aird.handlers.admin_handlers",
        "aird.handlers.api_handlers",
        "aird.handlers.auth_handlers",
        "aird.handlers.base_handler",
        "aird.handlers.file_op_handlers",
        "aird.handlers.share_handlers",
        "aird.handlers.view_handlers",
    ]
    modules = modules or default_modules
    for module in modules:
        targets.append(f"{module}.constants_module.DB_CONN")
        # Some modules reference DB_CONN directly
        targets.append(f"{module}.DB_CONN")

    patches = []
    active_patches = []
    try:
        for target in targets:
            patches.append(patch(target, value, create=True))
        for p in patches:
            try:
                p.start()
            except (AttributeError, ModuleNotFoundError):
                continue
            else:
                active_patches.append(p)
        yield value
    finally:
        for p in reversed(active_patches):
            p.stop()

